import difflib
import operator
import xml.dom.minidom
import xml.etree.ElementTree as ET
import re
import logging

from collections import defaultdict
from xml.dom import Node

# constants used for enums herein
CREATE, UPDATE, DELETE = "create", "update", "delete"
log = logging.getLogger("jjm")


class XmlChange:
    """Represents an xml job/view with current and new state."""

    __slots__ = ("name", "_before", "_after")

    sortable_node_names = {
        "project",
        "hudson.plugins.ws__cleanup.WsCleanup",
        "jenkins.plugins.slack.SlackNotifier",
    }

    def __init__(self, name: str):
        self.name = name
        if not name:
            raise ValueError("Name must be set")
        self._before = None
        self._after = None

    @staticmethod
    def xml_normalize(xml_str: str) -> str:
        """Normalize xml by running through a parser and removing blank text elements."""
        root = xml.dom.minidom.parseString(xml_str)
        # remove blanks
        node_queue = [root]
        while node_queue:
            node = node_queue.pop()

            if node.nodeName in XmlChange.sortable_node_names:
                # sort child elements for consistency
                node.childNodes.sort(key=operator.attrgetter("nodeName"))

            for n in node.childNodes:
                if n.nodeType is Node.TEXT_NODE and n.nodeValue:
                    n.nodeValue = n.nodeValue.strip()
                elif n.nodeType is Node.ELEMENT_NODE:
                    node_queue.append(n)

        root.normalize()
        return root.toprettyxml(indent="  ")

    def extract_md(self):
        node = ET.fromstring(self._after)
        desc = node.find("./description")
        if desc is None or not desc.text:
            log.warning("No description in jenkins job %r??", self.name)
            return {}
        text = desc.text.replace("<!-- Managed by Jenkins Job Builder -->", "")
        md = {
            m.group(1): m.group(2)
            for m in re.finditer(r"^\s*([\w-]+):\s*([\w -]+)\s*$", text, flags=re.M)
        }
        return md

    def changetype(self):
        if self._before == self._after:
            # No change
            ret = None
        elif not self._before and self._after:
            ret = CREATE
        elif self._before and not self._after:
            ret = DELETE
        else:
            ret = UPDATE
        return ret

    def difflines(self):
        difflines = difflib.unified_diff(
            (self._before or "").splitlines(),
            (self._after or "").splitlines(),
            fromfile=self.name,
            tofile=self.name,
        )
        return difflines

    @property
    def before_xml(self):
        return self._before

    @property
    def after_xml(self):
        return self._after

    @before_xml.setter
    def before_xml(self, val):
        self._before = self.xml_normalize(val)

    @after_xml.setter
    def after_xml(self, val):
        self._after = self.xml_normalize(val)


class XmlChangeDefaultDict(defaultdict):
    def __missing__(self, key):
        val = XmlChange(name=key)
        self[key] = val
        return val



import configparser
import hashlib
import logging
import os
import shlex

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("jjm")


class JenkinsConnectConfig:
    """
    Handle jenkins connection config.
    Includes methods for loading/storing per-user, per-url auth config.
    """

    global_conf_path = "/etc/jjm/jenkins_creds.ini"
    user_conf_path = os.path.expanduser("~/.config/jjm/jenkins_creds.ini")
    __slots__ = ("url", "username", "password", "timeout", "metadata", "allow_delete")

    def __init__(
        self, url, username, password, timeout=None, metadata=None, allow_delete=False
    ):
        if url is not None and url.endswith("/"):
            url = url[:-1]
        self.url = url
        self.username, self.password = username, password
        self.timeout = int(timeout or 60)
        self.metadata = metadata or MetadataConfig({})
        self.allow_delete = allow_delete

    def __str__(self):
        return (
            f"(url={self.url} username={self.username}"
            f" password={self.password_obscured} timeout={self.timeout})"
        )

    def __repr__(self):
        props = ",".join(f"{slot}={getattr(self, slot)!r}" for slot in self.__slots__)
        _repr = f"{self.__class__.__name__}({props})"
        return _repr

    @property
    def password_obscured(self):
        if self.password is None:
            return None
        return "md5:" + hashlib.md5(self.password.encode("utf-8")).hexdigest()

    @staticmethod
    def load_from_files(config_overrides=None):
        """loads config files in order, handles desired override structure"""
        cp = configparser.RawConfigParser()
        read_files = cp.read(
            [
                JenkinsConnectConfig.global_conf_path,
                JenkinsConnectConfig.user_conf_path,
                "./jjm.ini",
            ]
        )
        # cli overrides
        if config_overrides is not None:
            cp.read_dict({"jenkins": config_overrides})
        log.debug("loaded config: %r", read_files)
        url = cp.get("jenkins", "url", fallback=None)
        if not url:
            log.warning("Jenkins url not set.")
        elif url.endswith("/"):
            url = url[:-1]
        metadata = MetadataConfig.build_from_configparser(cp)

        def _section_by_url(key):
            return cp.get(url, key, fallback=None) or cp.get(
                "jenkins", key, fallback=None
            )

        loaded_config = JenkinsConnectConfig(
            url=url,
            username=_section_by_url("username"),
            password=_section_by_url("password"),
            timeout=_section_by_url("timeout"),
            metadata=metadata,
        )
        log.debug("loaded config=%r", loaded_config)
        return loaded_config

    def update_user_conf_auth(self, username, password):
        """create/update user config with auth values per url"""
        cp = configparser.RawConfigParser()
        cp.read([self.user_conf_path])
        if not cp.has_section(self.url):
            cp.add_section(self.url)
        cp.set(self.url, "username", username)
        cp.set(self.url, "password", password)
        os.makedirs(os.path.dirname(self.user_conf_path), exist_ok=True)
        with open(self.user_conf_path, "w") as fp:
            cp.write(fp)


class MetadataConfig:
    """
    Jenkins metadata plugin has been dead for a while,
    so this is a hack to the description field to deal with it.
    """

    __slots__ = ("metadata_conf", "required_fields", "valid_field_values")

    def __init__(self, metadata_conf: dict):
        self.metadata_conf = metadata_conf
        self.required_fields = (
            self.metadata_conf.get("required-description-fields") or []
        )
        self.valid_field_values = {}
        for field in self.required_fields:
            key = f"valid-values-for-{field}".lower()
            if key in self.metadata_conf:
                self.valid_field_values[field] = self.metadata_conf[key]
        log.debug("required_fields: %r", self.required_fields)
        log.debug("valid_field_values: %r", self.valid_field_values)

    @staticmethod
    def build_from_configparser(cp: configparser.RawConfigParser):
        if not cp.has_section("metadata"):
            return None
        md_conf = {}
        for key, val in cp.items(section="metadata"):
            if (
                key.startswith("valid-values-for-")
                or key == "required-description-fields"
            ):
                val = shlex.split(val)
            md_conf[key] = val
        return MetadataConfig(md_conf)

    def validate(self, md: dict):
        for field in self.required_fields:
            if field not in md:
                yield (
                    f"Missing metadata in job description: {field}."
                    f"\nAdd a line like `{field}: somevalue` to the description."
                )
        for field, field_values in self.valid_field_values.items():
            val = md.get(field)
            if val is None:
                continue
            if val not in field_values:
                yield (
                    f"Field {field} in job description has invalid value {val}."
                    f"\nValid values are {','.join(field_values)}"
                )



"""
Contains extensions for raw xml loading with jjb
"""
import os
import logging
import jinja2
import xml.etree.ElementTree as ET
import jenkins_jobs.modules.base
from jenkins_jobs.xml_config import XmlJob, XmlJobGenerator

log = logging.getLogger("jjm")


def load_xml_escaped(path):
    """shorthand function to work around some known issues with escaping from includes"""
    with open(path) as fp:
        contents = fp.read()
    return jinja2.filters.do_forceescape(contents)


# initialize a global jinja env
jenv = jinja2.Environment(
    loader=jinja2.FileSystemLoader(searchpath=(os.curdir,)),
    autoescape=True,
    undefined=jinja2.StrictUndefined,
)
jenv.globals["load_xml_escaped"] = load_xml_escaped


class RawXmlProject(jenkins_jobs.modules.base.Base):
    """add a job type for raw xml"""

    jenv = jenv

    def root_xml(self, data):
        name = data["name"]
        xmlstr = data["raw"]
        # add in a hack for raw jobs to do optional jinja templating
        if data.get("jinja") is True:
            log.debug("Rendering jinja template for %s", name)
            templ = self.jenv.from_string(xmlstr)
            xmlstr = templ.render(data=data)
            log.debug("rendered: \n---\n%s\n---", xmlstr)
        xml_parent = ET.fromstring(xmlstr)
        return xml_parent


class XmlJobGeneratorWithRaw(XmlJobGenerator):
    """bypasses the module loader for the raw xml job type"""

    def _annotate_with_plugins(self, xml_job: XmlJob):
        """Many elements coming out of jjb are missing plugin version data."""
        plugins: dict = self.registry.plugins_dict
        doc: ET.Element = xml_job.xml
        for node in doc.iterfind(".//*[@plugin]"):
            plugin_name = node.attrib["plugin"]
            if "@" in plugin_name:
                continue
            version = plugins[plugin_name]["version"]
            log.debug("annotated %r with %s@%s", node, plugin_name, version)
            node.attrib["plugin"] = f"{plugin_name}@{version}"

    def _getXMLForData(self, data):
        kind = data.get(self.kind_attribute, self.kind_default)
        if kind == "raw":
            mod = RawXmlProject(self.registry)
            _xml = mod.root_xml(data)
            obj = XmlJob(_xml, data["name"])
            return obj
        xml_job = super(XmlJobGenerator, self)._getXMLForData(data)
        # self._annotate_with_plugins(xml_job)
        return xml_job



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jenkins_job_manager']

package_data = \
{'': ['*'], 'jenkins_job_manager': ['j2_templates/*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'jenkins-job-builder>=3.5.0,<4.0.0',
 'jinja2>=2.11,<3.0',
 'python-jenkins>=1.7.0,<2.0.0']

entry_points = \
{'console_scripts': ['jjm = jenkins_job_manager.cli:jjm']}

setup_kwargs = {
    'name': 'jenkins-job-manager',
    'version': '0.6.1',
    'description': 'A terraform-like wrapper around JJB with some additional features.',
    'long_description': None,
    'author': 'Jeremy Lavergne',
    'author_email': 'github@lavergne.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
"""
Wrapper tool for managing jenkins jobs via jenkins job builder
"""
from jenkins_job_manager.core import JenkinsJobManager
from jenkins_job_manager.xml_change import DELETE

import click
import logging
import os
import typing

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("jjm")

HERE = os.path.dirname(os.path.realpath(__file__))


click_option_target = click.option(
    "--target",
    default=None,
    multiple=True,
    help="job name glob to target",
)


@click.group()
@click.option("--debug", "-d", is_flag=True)
@click.option("--working-dir", "-C", default=None, help="change to this directory ")
@click.option("--url", help="jenkins base url")
@click.pass_context
def jjm(ctx, debug, working_dir, url):
    """Jenkins Job Management"""
    if debug:
        log.setLevel(logging.DEBUG)
    if working_dir:
        os.chdir(working_dir)

    config = {}
    if url:
        config["url"] = url
    jjm = JenkinsJobManager(config_overrides=config)
    ctx.obj = jjm
    if not jjm.config.url:
        click.echo(
            "\n"
            "ERROR: No jenkins url configured.\n"
            "Create a ./jjm.ini file with contents:\n"
            "    [jenkins]\n"
            "    url = https://yourjenkinsurl.com/\n"
        )
        raise click.exceptions.Exit(1)


@jjm.command(name="login")
@click.pass_obj
def jjm_login(obj: JenkinsJobManager):
    """store login config per url"""
    jjm = obj
    jconf = jjm.config
    jurl = jconf.url
    username, password = jconf.username, jconf.password
    if username and password:
        click.secho("Auth already configured for this jenkins!", fg="red")
        click.secho(f"{jconf}", fg="white")
        click.confirm("overwrite?", abort=True)

    click.secho("Configuring login info for:", fg="green")
    click.secho(f"\t{jurl}", fg="white")

    click.secho(
        f"\nEnter username, If unsure go to {jurl}/whoAmI/"
        f" (if this says anonymous, you need to login first)"
    )
    username = click.prompt("username", type=str)

    click.secho(
        f"\nEnter api key. go to {jurl}/user/{username}/configure to make a new one."
    )
    password = click.prompt("api key", type=str, hide_input=True)
    log.debug("entered username=%r password=%r", username, password)

    jconf.username, jconf.password = username, password
    if jjm.check_authentication():
        click.secho(f"Success! Saving to {jconf.user_conf_path}", fg="green")
        jconf.update_user_conf_auth(username, password)
    else:
        click.secho("Bad Authentication, try again.", fg="red")
        raise click.exceptions.Exit(2)


def handle_validation_errors(obj: JenkinsJobManager, ignore=False):
    warnings = obj.validate_metadata()
    errors = False
    for job_name, warning in warnings:
        if errors is False:
            click.secho("Validation Errors", fg="red", bold=True)
            errors = True
        click.secho(f"Job {job_name!r}: {warning}", fg="red")

    if errors is True and ignore is False:
        raise click.exceptions.Exit(5)


@jjm.command(name="check")
@click.option("--load-plugins", is_flag=True)
@click.pass_obj
def jjm_check(obj: JenkinsJobManager, load_plugins):
    """check syntax/config"""
    if load_plugins:
        obj.load_plugins_list()
    obj.generate_jjb_xml()
    handle_validation_errors(obj)


def check_auth(obj: JenkinsJobManager):
    """cli helper for auth check"""
    if not obj.check_authentication():
        click.secho(f"Bad login detected for {obj.config}", fg="red")
        click.echo("Try the login subcommand")
        raise click.exceptions.Exit(1)


def handle_plan_report(obj: JenkinsJobManager, use_pager=True, output=None) -> bool:
    """cli helper for plan report"""
    changes = obj.detected_changes()

    if output:
        # machine readable
        for line in obj.plan_report(report_format=output):
            click.echo(line, nl=False)
    elif changes:
        # default "human" readable, show changes
        def output_format(line):
            if line.startswith("+"):
                return click.style(line, fg="green")
            elif line.startswith("-"):
                return click.style(line, fg="red")
            else:
                return line

        gen_lines = map(output_format, obj.plan_report())
        if use_pager is True:
            click.echo_via_pager(gen_lines)
        else:
            for line in gen_lines:
                click.echo(line, nl=False)
    else:
        # default "human" readable, no changes
        click.secho("No changes.", fg="green")

    return changes


@jjm.command(name="plan")
@click.option("--skip-pager", is_flag=True)
@click.option(
    "--output", default=None, type=click.Choice(["json", "yaml"], case_sensitive=False)
)
@click_option_target
@click.pass_obj
def jjm_plan(
    obj: JenkinsJobManager,
    skip_pager: bool,
    output: typing.Optional[str],
    target: typing.List[str],
):
    """check for changes"""
    check_auth(obj)
    obj.gather(target)
    handle_validation_errors(obj)
    changes = handle_plan_report(obj, use_pager=not skip_pager, output=output)
    if changes is True:
        raise click.exceptions.Exit(2)


@jjm.command(name="apply")
@click.option("--allow-delete", is_flag=True, help="disables delete safety.")
@click_option_target
@click.pass_obj
def jjm_apply(obj: JenkinsJobManager, target: str, allow_delete: bool):
    """check and apply changes"""
    if allow_delete:
        obj.config.allow_delete = True
    check_auth(obj)
    obj.gather(target)
    handle_validation_errors(obj)
    if obj.detected_changes() is False:
        click.secho("No changes to apply.", fg="green")
        return
    handle_plan_report(obj, use_pager=False)
    click.confirm(click.style("Apply changes?", bold=True), abort=True)
    changecounts, msg = obj.apply_plan()
    click.echo(msg)
    _deletes = changecounts[DELETE]
    if allow_delete is False and _deletes > 0:
        click.secho(
            f"Ignored {_deletes} deletes.\nUse --allow-delete to override.",
            fg="red",
        )


@jjm.command(name="import")
@click_option_target
@click.pass_obj
def jjm_import(obj: JenkinsJobManager, target: str):
    check_auth(obj)
    obj.gather(target)
    missing = obj.import_missing()
    click.secho(f"Imported {len(missing)} jobs.", fg="green")


if __name__ == "__main__":
    jjm()



from jenkins_job_manager.connect_config import JenkinsConnectConfig
from jenkins_job_manager.xml_change import (
    XmlChange,
    XmlChangeDefaultDict,
    CREATE,
    UPDATE,
    DELETE,
)
from jenkins_job_manager.raw_ext import (
    RawXmlProject,
    XmlJobGeneratorWithRaw,
)

import fnmatch
import glob
import itertools
import logging
import os
import random
import re
import string
import xml.etree.ElementTree as ET
from typing import Dict, Optional

import jenkins
import jinja2
from jenkins_jobs.parser import YamlParser
from jenkins_jobs.registry import ModuleRegistry
from jenkins_jobs.xml_config import XmlJob, XmlViewGenerator

HERE = os.path.dirname(os.path.realpath(__file__))
J2_DIR = f"{HERE}/j2_templates"

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("jjm")


class NameRegexFilter:
    """Regex Filter Callable"""

    __slots__ = ("regex",)

    def __init__(self, regexp):
        self.regex = re.compile(regexp)

    @staticmethod
    def from_glob_list(globs):
        regexes = map(fnmatch.translate, globs)
        combined_regex = "|".join(regexes)
        return NameRegexFilter(combined_regex)

    def __call__(self, job_name):
        m = self.regex.match(job_name)
        return True if m is not None else False

    def __repr__(self):
        return f"{self.__class__.__name__}<{repr(self.regex)[11:-1]}>"


class JenkinsJobManager:
    """main jjb manager"""

    __slots__ = (
        "config",
        "plugins_list",
        "_jenkins",
        "jobs",
        "_jobs_filter_func",
        "views",
        "validation_errors",
        "jenv",
    )
    job_managing_job_classes = frozenset(["jenkins.branch.OrganizationFolder"])
    raw_xml_yaml_path = "./raw_xml_jobs.yaml"

    def __init__(self, config_overrides=None):
        self.config: JenkinsConnectConfig = JenkinsConnectConfig.load_from_files(
            config_overrides
        )
        self._jenkins: Optional[jenkins.Jenkins] = None
        self.plugins_list: Optional[list] = None
        self.jobs: Dict[str, XmlChange] = XmlChangeDefaultDict()
        self._jobs_filter_func: NameRegexFilter = NameRegexFilter(".*")
        self.views: Dict[str, XmlChange] = XmlChangeDefaultDict()
        self.validation_errors = []
        self.jenv = jinja2.Environment(
            loader=jinja2.FileSystemLoader([J2_DIR]),
            undefined=jinja2.StrictUndefined,
            autoescape=False,
        )

    @property
    def jenkins(self):
        if self._jenkins is None:
            self._jenkins = jenkins.Jenkins(
                url=self.config.url,
                username=self.config.username,
                password=self.config.password,
                timeout=self.config.timeout,
            )
        return self._jenkins

    def check_authentication(self):
        """check if jenkins connection config correct"""
        log.debug("checking credentials")
        try:
            result = self.jenkins.get_whoami()
            if self.config.username != result["id"]:
                raise RuntimeError(f"{self.config.username!r} != {result['id']!r}")
        except (jenkins.BadHTTPException, jenkins.JenkinsException) as e:
            log.debug("BadHTTPException: %s", e)
            return False
        return True

    def read_views(self):
        """read existing views from jenkins"""
        jenkins = self.jenkins
        views = self.views
        log.info("Reading jenkins views state")

        for view_d in jenkins.get_views():
            log.debug("found view %r", view_d)
            name, url, _class = view_d["name"], view_d["url"], view_d.get("_class")
            if not self._jobs_filter_func(name):
                log.debug("Ignored by filter: %s", name)
                continue
            if name == "All" or name == "all" or _class == "hudson.model.AllView":
                log.debug("ignoring AllView: %r", view_d)
                continue
            view_config = jenkins.get_view_config(name)
            views[name].before_xml = view_config

    def read_jobs(self):
        """read existing jobs from jenkins"""
        jenkins = self.jenkins
        jobs = self.jobs

        log.info("Reading jenkins jobs state")
        # ignorable subfolder jobs
        managed_job_urls = set()
        _empty = tuple()
        for d in jenkins.get_all_jobs():
            log.debug("found job %r", d)
            name, url, _class = d["fullname"], d["url"], d.get("_class")
            subjobs = d.get("jobs", _empty)
            if not self._jobs_filter_func(name):
                log.debug("Ignored by filter: %s", name)
                continue
            elif _class in self.job_managing_job_classes:
                managed_job_urls.update(job_d["url"] for job_d in subjobs)
            elif url in managed_job_urls:
                log.debug("Ignoring managed job %s", name)
                # recursively ignore jobs of ignored jobs
                managed_job_urls.update(job_d["url"] for job_d in subjobs)
                continue
            job_conf = jenkins.get_job_config(name)
            jobs[name].before_xml = job_conf

    def load_plugins_list(self):
        """load plugin info in format expected by jjb libs"""
        log.debug("reading jenkins plugins")
        self.plugins_list = list(self.jenkins.get_plugins().values())

    @staticmethod
    def xml_dump(root: ET.Element) -> str:
        return ET.tostring(root, encoding="unicode")

    def get_jjb_config(self):
        class JJBConfig:
            yamlparser = {
                "allow_duplicates": False,
                "keep_descriptions": True,
                "include_path": ".",
                "retain_anchors": None,
            }

            @staticmethod
            def get_plugin_config(*args, **kwargs):
                return None

        return JJBConfig

    def jenkins_format_xml(self, xml_job: XmlJob):
        """
        bounces job config through jenkins to get the formatting right
        unused, deprecated
        """
        jenkins = self.jenkins
        _xml: ET.Element = xml_job.xml

        if not _xml.find("./disabled"):
            d = ET.Element("disabled")
            d.text = "false"
            _xml.append(d)
        disabled_job = _xml.find("./disabled").text == "true"

        rand_suffix = "".join(random.choice(string.hexdigits) for _ in range(10))
        tmp_name = f"zz_jjm_tmp_{xml_job.name}_{rand_suffix}"
        tmp_xml = xml_job.xml
        tmp_xml.find("./disabled").text = "true"
        tmp_xml_str = ET.tostring(tmp_xml, encoding="unicode")
        log.info("creating %s", tmp_name)
        jenkins.create_job(tmp_name, tmp_xml_str)
        formatted_xml = jenkins.get_job_config(tmp_name)
        log.info("removing %s", tmp_name)
        jenkins.delete_job(tmp_name)
        if not disabled_job:
            formatted_xml = formatted_xml.replace(
                "<disabled>true</disabled>", "<disabled>false</disabled>"
            )
        return formatted_xml

    def generate_jjb_xml(self):
        """render jjb yaml to xml"""

        jjb_config = self.get_jjb_config()
        options_names = []  # normally a list of jobs globs for targeting
        files_path = glob.glob("./**/", recursive=True)

        parser = YamlParser(jjb_config)
        registry = ModuleRegistry(jjb_config, self.plugins_list)

        xml_job_generator = XmlJobGeneratorWithRaw(registry)
        xml_view_generator = XmlViewGenerator(registry)

        parser.load_files(files_path)
        registry.set_parser_data(parser.data)

        job_data_list, view_data_list = parser.expandYaml(registry, options_names)

        def job_data_filter_wrapper(job_data):
            return self._jobs_filter_func(job_data["name"])

        xml_jobs = xml_job_generator.generateXML(
            filter(job_data_filter_wrapper, job_data_list)
        )
        jobs = self.jobs
        for xml_job in xml_jobs:
            formatted_xml_str = self.xml_dump(xml_job.xml)
            jobs[xml_job.name].after_xml = formatted_xml_str

        xml_views = xml_view_generator.generateXML(
            filter(job_data_filter_wrapper, view_data_list)
        )
        views = self.views
        for xml_view in xml_views:
            views[xml_view.name].after_xml = self.xml_dump(xml_view.xml)

    def detected_changes(self):
        return any(
            item.changetype() is not None
            for item in itertools.chain(self.jobs.values(), self.views.values())
        )

    def gather(self, target_job_names=None):
        """run this to gather plan/apply data"""
        if target_job_names:
            log.debug("target_job_names=%r", target_job_names)
            self._jobs_filter_func = NameRegexFilter.from_glob_list(target_job_names)
        self.read_views()
        self.read_jobs()
        self.load_plugins_list()
        self.generate_jjb_xml()

    def import_missing(self) -> list:
        """import missing jobs as xml"""
        missing = [item for item in self.jobs.values() if item.changetype() is DELETE]
        if not missing:
            return []

        class FakeRegistry:
            modules = []

        def job_name_to_file_name(j_name):
            _part = re.sub(r"[\/]", "_", j_name)
            return f"./{_part}.xml"

        xml_job_name_pairs = []

        if os.path.exists(self.raw_xml_yaml_path):
            parser = YamlParser(self.get_jjb_config())
            parser.load_files([self.raw_xml_yaml_path])
            job_data_list, _ = parser.expandYaml(FakeRegistry, [])
            for job_data in job_data_list:
                name = job_data["name"]
                fname = job_name_to_file_name(name)
                assert os.path.exists(fname)
                xml_job_name_pairs.append((name, fname))
        template = self.jenv.get_template("raw_xml_import.j2")

        for mxml in missing:
            job_name = mxml.name
            file_name = job_name_to_file_name(job_name)
            job_config = mxml.before_xml
            xml_job_name_pairs.append((job_name, file_name))
            assert not os.path.exists(file_name)
            with open(file_name, "w") as fp:
                fp.write(job_config)
            log.info("Imported %s to %s", job_name, file_name)

        with open(self.raw_xml_yaml_path, "w") as fp:
            template.stream(raw_xml_jobs=xml_job_name_pairs).dump(fp)
        return missing

    def validate_metadata(self):
        md_conf = self.config.metadata

        for job in self.jobs.values():
            if job.after_xml is None:
                continue
            md = job.extract_md()
            warnings = md_conf.validate(md)
            for warning in warnings:
                yield job.name, warning

    def plan_report(self, report_format=None):
        """report on changes about to be made"""
        if report_format == "json":
            template_name = "json.j2"
        elif report_format == "yaml":
            template_name = "yaml.j2"
        else:
            template_name = "default.j2"
        template = self.jenv.get_template(template_name)

        changecounts = {CREATE: [], UPDATE: [], DELETE: []}

        def iter_changes(xml_dict, output=None):
            """closure to handle changecount side effect"""

            for item in xml_dict.values():
                changetype = item.changetype()
                if changetype is None:
                    continue
                if output:
                    md = item.extract_md() or {}
                    yield item.name, item.before_xml, item.after_xml, item.difflines(), md, item.changetype()
                else:
                    for i, line in enumerate(item.difflines()):
                        # deals with the rare case that the diff shows no lines
                        if i == 0:
                            changecounts[changetype].append(item.name)
                        yield line

        report_context = {
            "view_changes": iter_changes(self.views),
            "job_changes": iter_changes(self.jobs, report_format),
            "changecounts": changecounts,
        }

        return template.generate(
            obj=report_context, CREATE=CREATE, UPDATE=UPDATE, DELETE=DELETE
        )

    def apply_plan(self):
        """apply changes from gather/plan"""
        changecounts = {CREATE: 0, UPDATE: 0, DELETE: 0}
        log.debug("applying views")
        for view in self.views.values():
            changetype = view.changetype()
            if changetype is None:
                log.debug("no change: %s", view.name)
                continue
            elif changetype is CREATE:
                log.info("create view %s", view.name)
                self.jenkins.create_view(view.name, view.after_xml)
                changecounts[CREATE] += 1
            elif changetype is UPDATE:
                log.info("reconfig view %s", view.name)
                self.jenkins.reconfig_view(view.name, view.after_xml)
                changecounts[UPDATE] += 1
            elif changetype is DELETE:
                log.info("delete view %s", view.name)
                if self.config.allow_delete:
                    self.jenkins.delete_view(view.name)
                else:
                    log.warning("refusing to delete view %s", view.name)
                changecounts[DELETE] += 1
            else:
                raise RuntimeError(
                    f"Invalid changetype {changetype}(id={id(changetype)})"
                )
        log.debug("applying jobs")
        for job in self.jobs.values():
            changetype = job.changetype()
            if changetype is None:
                log.debug("no change: %s", job.name)
                continue
            elif changetype is CREATE:
                log.info("create job %s", job.name)
                self.jenkins.create_job(job.name, job.after_xml)
                changecounts[CREATE] += 1
            elif changetype is UPDATE:
                log.info("reconfig job %s", job.name)
                self.jenkins.reconfig_job(job.name, job.after_xml)
                changecounts[UPDATE] += 1
            elif changetype is DELETE:
                log.info("delete job %s", job.name)
                if self.config.allow_delete:
                    self.jenkins.delete_job(job.name)
                else:
                    log.warning("refusing to delete job %s", job.name)
                changecounts[DELETE] += 1
            else:
                raise RuntimeError(
                    f"Invalid changetype {changetype}(id={id(changetype)})"
                )
        msg = (
            f"Changes applied. added={changecounts[CREATE]} updated={changecounts[UPDATE]}"
            f" deleted={changecounts[DELETE]}"
        )
        return changecounts, msg
